/* This file is part of dHole client-server package.
 *
 * (c) 2002 Dark0n3
 *
 */
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <resolv.h>

#define MAX_CLIENTS		1024
#define BUF_SIZE		4096
#define PROXY			"127.0.0.1"
#define PROXY_PORT		9999
#define SERVER			"127.0.0.1"
#define SERVER_PORT		22222
#define RETRY			600
#define LOGFILE			"client.log"
#define LOG_CONNECTIONS		1
#define LOG_ERRORS		1
#define LOG_DEBUG		0



#if ( LOG_ERRORS )
#define E_LOG	log
#else
#define E_LOG
#endif
#if ( LOG_CONNECTIONS )
#define C_LOG	log
#else
#define C_LOG
#endif
#if ( LOG_DEBUG )
#define D_LOG	log
#else
#define D_LOG
#endif


/*
 * void log(char *format, ...);
 *
 * Appends stuff to LOGFILE
 *
 */
void log(char *fmt, ...) {
 time_t		timenow;
 FILE		*file;
 va_list	ap;

 va_start(ap, fmt);
 timenow = time(NULL);

 if ((file = fopen(LOGFILE, "a+")) != NULL ) {
	fprintf(file, "%.24s - ", ctime(&timenow));
	vfprintf(file, fmt, ap);
	fclose(file);
	}
}




main (int argc, char **argv) {
 struct		sockaddr_in	remote_addr;
 struct		hostent		*hp;
 unsigned char			buf[BUF_SIZE];
 fd_set				rfds;
 int				client_s[2][MAX_CLIENTS];
 int				control_s;
 int				sz;
 int				n, m;

 bzero(&client_s, MAX_CLIENTS * sizeof(int) * 2);
 sz = sizeof(struct sockaddr_in);
 control_s = 0;

#if ( LOG_CONNECTIONS || LOG_ERRORS || LOG_DEBUG )
 log("New session started\n");
#endif

 if ( fork() ) {
	exit(0);
	}

 D_LOG("Succesfully forked main process to background\n");

 while ( 1 ) {
	if ( control_s == 0 ) { /* Infinite loop until control connection is established */
		bzero(&remote_addr, sz);

		remote_addr.sin_family = AF_INET;
		remote_addr.sin_port = htons((unsigned)PROXY_PORT);

		hp = gethostbyname((char *)PROXY);
		memcpy(&remote_addr.sin_addr, hp->h_addr, hp->h_length );

		while ( 1 ) {
			C_LOG("Connecting to proxy server [control]\n");
			control_s = socket(AF_INET, SOCK_STREAM, 0);
			if ( connect(control_s, (struct sockaddr *)&remote_addr, sz) == 0 ) {
				break;
				}
			C_LOG("Could not reach the proxy server, retrying in %isecs\n", RETRY);
			close(control_s);
			sleep(RETRY);
			}
		D_LOG("Sending hello to server\n");
		n = -1;
		write(control_s, &n, sizeof(int));
		}

	FD_ZERO(&rfds);
	FD_SET(control_s, &rfds);
	m = control_s;

	for (n = 0; n < MAX_CLIENTS; n++) if (client_s[0][n] > 0 && client_s[1][n] > 0) {
		FD_SET(client_s[0][n], &rfds);
		FD_SET(client_s[1][n], &rfds);
		m = m > client_s[0][n] ? m : client_s[0][n];
		m = m > client_s[1][n] ? m : client_s[1][n];
		}
	m++;
	select(m, &rfds, NULL, NULL, NULL);

	if ( FD_ISSET(control_s, &rfds) ) {	/* Data coming from proxy */
		m = read(control_s, &n, sizeof(int));
		if ( m > 0 && n >= 0 ) {	/* Proxy requested data connection */
			C_LOG("Proxy server requested new data connection\n");
			bzero(&remote_addr, sz);

			remote_addr.sin_family = AF_INET;
			remote_addr.sin_port = htons((unsigned)SERVER_PORT);
			hp = gethostbyname((char *)SERVER);

			D_LOG("Connecting to server [data #%i]\n", n);

			memcpy(&remote_addr.sin_addr, hp->h_addr, hp->h_length);
			client_s[1][n] = socket(AF_INET, SOCK_STREAM, 0);
			connect(client_s[1][n], (struct sockaddr *)&remote_addr, sz);

			bzero(&remote_addr, sz);
			remote_addr.sin_family = AF_INET;
			remote_addr.sin_port = htons((unsigned)PROXY_PORT);
			hp = gethostbyname((char *)PROXY);

			D_LOG("Connecting to proxy server [data #%i]\n", n);
			memcpy(&remote_addr.sin_addr, hp->h_addr, hp->h_length);
			client_s[0][n] = socket(AF_INET, SOCK_STREAM, 0);
			connect(client_s[0][n], (struct sockaddr *)&remote_addr, sz);

			D_LOG("Validating new data connection [data #%i]\n", n);
			write(client_s[0][n], &n, sizeof(int)); /* Currently we dont send info wheter connection failed or not */
			} else {	/* Control connection dead, wipe all active data conections */
			C_LOG("Connection to proxy server lost\n");
			for (n = 0; n < MAX_CLIENTS; n++) {
				if ( client_s[0][n] > 0 ) {
					close(client_s[0][n]);
					}
				if ( client_s[1][n] > 0 ) {
					close(client_s[1][n]);
					}
				}
			bzero(&client_s, MAX_CLIENTS * sizeof(int) * 2);
			close(control_s);
			control_s = 0;
			}
		}

	for (n = 0; n < MAX_CLIENTS; n++) if (client_s[0][n] > 0 && client_s[1][n] > 0){
		if ( FD_ISSET(client_s[0][n], &rfds) ) { /* Data from proxy */
			D_LOG("Data available in proxy server socket [data #%i]\n", n);
			if ((m = read(client_s[0][n], buf, BUF_SIZE)) <= 0 ||
			    write(client_s[1][n], buf, m) != m) {
				E_LOG("Data connection closed [data #%i]\n", n);
				close(client_s[0][n]);
				close(client_s[1][n]);
				client_s[0][n] = 0;
				client_s[1][n] = 0;
				continue;
				}
			D_LOG("Succesfully forwarded data from proxy server to server [#%i]\n", n);
			}
		if ( FD_ISSET(client_s[1][n], &rfds) ) { /* Data from server */
			D_LOG("Data available in server socket [#%i]\n", n);
			if ((m = read(client_s[1][n], buf, BUF_SIZE)) <= 0 ||
			    write(client_s[0][n], buf, m) != m) {
				E_LOG("Data connection closed [data #%i]\n", n);
				close(client_s[0][n]);
				close(client_s[1][n]);
				client_s[0][n] = 0;
				client_s[1][n] = 0;
				continue;
				}
			D_LOG("Succesfully forwarded data from server to proxy server [#%i]\n", n);
			}
		}
	}
}
