/* This file is part of dHole client-server application.
 *
 * (c) 2002 Dark0n3
 *
 */
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <resolv.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <resolv.h>

#define CLIENT_PORT		9998
#define PROXY_PORT		9999
#define	MAX_CLIENTS		1024
#define BUF_SIZE		4096
#define LOGFILE			"server.log"
#define LOG_CONNECTIONS		1
#define LOG_ERRORS		1
#define LOG_DEBUG		0


#if ( LOG_ERRORS )
#define E_LOG	log
#else
#define E_LOG
#endif
#if ( LOG_CONNECTIONS )
#define C_LOG	log
#else
#define C_LOG
#endif
#if ( LOG_DEBUG )
#define D_LOG	log
#else
#define D_LOG
#endif


/*
 * void log(char *format, ...);
 *
 * Appends stuff to LOGFILE
 *
 */
void log(char *fmt, ...) {
 time_t         timenow;
 FILE           *file;
 va_list        ap;

 va_start(ap, fmt);
 timenow = time(NULL);

 if ((file = fopen(LOGFILE, "a+")) != NULL ) {
	fprintf(file, "%.24s - ", ctime(&timenow));
	vfprintf(file, fmt, ap);
	fclose(file);
	}
}



main (int argc, char **argv) {
 struct		sockaddr_in	remote_addr;
 struct		sockaddr_in	local_addr;
 fd_set				rfds;
 unsigned char			buf[BUF_SIZE];
 int				client_s[2][1024]; /* 0 for client-user, 1 for client-proxy */
 int				listen_s[2]; /* 0 client-user, 1 client-proxy */
 int				control_s;
 int				m;
 int				n;
 int				sz;
 unsigned			port[2];

 port[0] = (unsigned)CLIENT_PORT;
 port[1] = (unsigned)PROXY_PORT;

 control_s = -1;

 sz = sizeof(struct sockaddr_in);
 bzero(&client_s, MAX_CLIENTS * sizeof(int) * 2);

 printf("dHole Proxy Server			dark0n3 (c) 2002\n");

 for (n = 0; n < 2; n++) {
	bzero(&local_addr, sz);
	local_addr.sin_addr.s_addr = INADDR_ANY;
	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons((unsigned)port[n]);
	listen_s[n] = socket(PF_INET, SOCK_STREAM, 0);
	while ( bind(listen_s[n], (struct sockaddr *)&local_addr, sz) != 0 ||
		listen(listen_s[n], MAX_CLIENTS) != 0 ) {
		local_addr.sin_port = htons((unsigned)port[n]);

		printf("Could not bind to port: %u - waiting 30seconds\n", port[n]);
		sleep(30);
		}
	}


#if ( LOG_CONNECTIONS || LOG_ERRORS || LOG_DEBUG )
 log("New session started\n");
#endif

 if ( fork() ) {
	exit(0);
	}

 D_LOG("Succesfully forked main process to background\n");

 while ( 1 ) {
	FD_ZERO(&rfds);
	FD_SET(listen_s[1], &rfds);

	if (control_s > 0) {
		FD_SET(listen_s[0], &rfds);
		m = listen_s[0] > listen_s[1] ? listen_s[0] : listen_s[1];

		FD_SET(control_s, &rfds);
		m = m > control_s ? m : control_s;
		} else {
		m = listen_s[1];
		}

	for (n = 0; n < MAX_CLIENTS; n++ ) if ( control_s > 0 && client_s[0][n] > 0 && client_s[1][n] > 0 ) {
		FD_SET(client_s[0][n], &rfds);
		FD_SET(client_s[1][n], &rfds);
		m = m > client_s[0][n] ? m : client_s[0][n];
		m = m > client_s[1][n] ? m : client_s[1][n];
		}
	m++;
	select(m, &rfds, NULL, NULL, NULL); /* wait until data arrives */


	if ( control_s > 0 ) {
		for (n = 0; n < MAX_CLIENTS; n++ ) if ( client_s[0][n] > 0 && client_s[1][n] > 0 ) {
			if ( FD_ISSET(client_s[0][n], &rfds) ) {
				D_LOG("Data available in client socket [#%i]\n", n);
				if ((m = read(client_s[0][n], buf, BUF_SIZE)) <= 0 ||
				    write(client_s[1][n], buf, m) != m) { /* Connection closed */
					E_LOG("Data connection closed [#%i]\n", n);
					close(client_s[0][n]);
					close(client_s[1][n]);
					client_s[0][n] = 0;
					client_s[1][n] = 0;
					continue;
					}
				D_LOG("Succesfully forwarded data from client to proxy client [#%i]\n", n);
				}
			if ( FD_ISSET(client_s[1][n], &rfds) ) {
				D_LOG("Data available in proxy client socket [#%i]\n", n);
				if ((m = read(client_s[1][n], buf, BUF_SIZE)) <= 0 ||
				    write(client_s[0][n], buf, m) != m) { /* Connection closed */
					E_LOG("Data connection closed [#%i]\n", n);
					close(client_s[0][n]);
					close(client_s[1][n]);
					client_s[0][n] = 0;
					client_s[1][n] = 0;
					continue;
					}
				D_LOG("Succesfully forwarded data from proxy client to client [#%i]\n", n);
				}
			}

		if ( FD_ISSET(listen_s[0], &rfds)) {	/* new connection from client-user */
			bzero(&remote_addr, sz);
			for (n = 0; n < MAX_CLIENTS; n++) {
				if ( client_s[0][n] > 0 ) continue;
				client_s[0][n] = accept(listen_s[0], (struct sockaddr *)&remote_addr, &sz);
				C_LOG("New client connected\n");
				client_s[1][n] = 0;
				D_LOG("Requesting proxy client to establish data connection\n");
				write(control_s, &n, sizeof(int));
				m = 1;
				break;
				}
			}

		if ( FD_ISSET(control_s, &rfds) ) { /* Client shouldnt send anything through this socket - disconnect it */
			C_LOG("Proxy client disconnected\n");
			close(control_s);
			control_s = -1;
			for (n = 0; n < MAX_CLIENTS; n++) {
				if ( client_s[0][n] > 0 ) {
					close(client_s[0][n]);
					client_s[0][n] = 0;
					}
				if ( client_s[1][n] > 0 ) {
					close(client_s[1][n]);
					client_s[1][n] = 0;
					}
				}
			D_LOG("Succesfully closed all open connections\n");
			}
		}

	if ( FD_ISSET(listen_s[1], &rfds) ) {	/* data coming from client-proxy */
		bzero(&remote_addr, sz);
		m = accept(listen_s[1], (struct sockaddr *)&remote_addr, &sz);
		read(m, &n, sizeof(int));
		if ( n == -1 ) {
			if ( control_s == -1 ) {
				C_LOG("Proxy client connected\n");
				control_s = m;
				} else {
				E_LOG("Proxy client tried to connect, but control connection already exists\n");
				close(m); /* die, we dont need two servers */
				}
			} else {
			if ( n >= MAX_CLIENTS || n < 0 || client_s[1][n] != 0 ) {
				E_LOG("Unknown response from connecting client\n");
				close(m); /* this wasnt real data */
				} else {
				D_LOG("Proxy client data connection succesfully established\n");
				client_s[1][n] = m;
				}
			}
		}
	}
}
